# -*- coding: utf-8 -*-

from .basic_profile import BasicProfile


class RadialProfile(BasicProfile):
    def __init__(self, image_holder, parent=None):
        self.image_holder = image_holder
        super().__init__(parent)

    def update_data_from_source(self):
        image_holder = self.image_holder
        profile = image_holder.get_radial_profile()
        if profile is None:
            return
        r_axis = image_holder.geometry.r_axis
        if r_axis is not None and r_axis.size == profile.size:
            self.set_data(profile, r_axis)



import sys
import logging

from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import Qt

from .app import App
from .gui import GIWAXSMainWindow, UncaughtHook


def run():
    logging.basicConfig(level=logging.ERROR)
    exception_hook = UncaughtHook()
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    qapp = QApplication(sys.argv)
    giwaxs_app = App()
    window = GIWAXSMainWindow()
    sys.exit(qapp.exec_())



# -*- coding: utf-8 -*-

from dataclasses import dataclass
from typing import Tuple, Dict

import numpy as np
from scipy.optimize import curve_fit

from .rois.roi import Roi, RoiTypes
from .file_manager import ImageKey


@dataclass
class Fit:
    roi: Roi
    image_key: ImageKey
    r_range: Tuple[float, float]
    x_range: Tuple[int, int]
    x: np.ndarray
    y: np.ndarray
    init_curve: np.ndarray
    init_params: list
    lower_bounds: list
    upper_bounds: list
    fitted_params: list
    fit_errors: list
    fitting_curve: np.ndarray
    param_names: list
    fitting_method: str

    @property
    def bounds(self):
        return tuple(self.lower_bounds), tuple(self.upper_bounds)

    def update(self, **params):
        for k, v in params.items():
            setattr(self, k, v)


class GaussianFit(object):
    PARAM_NAMES = ('peak height', 'radius', 'width', 'background')
    METHOD = 'Gaussian'
    PADDING = 0.7

    def __init__(self, image_key: ImageKey, polar_image: np.ndarray, r_axis: np.ndarray, phi_axis: np.ndarray):
        self.image_key = image_key
        self.polar_image = polar_image
        self.r_axis = r_axis
        self.phi_axis = phi_axis
        self.r_delta = (r_axis.max() - r_axis.min()) / r_axis.size
        self.phi_delta = (phi_axis.max() - phi_axis.min()) / phi_axis.size
        self.r_profile = polar_image.sum(axis=0)
        self.fits: Dict[int, Fit] = {}
        self.aspect_ratio = self._aspect_ratio()
        self.bounds = self._bounds()
        self.name = ''

    def _aspect_ratio(self):
        p, r = self.phi_axis, self.r_axis
        return (p.max() - p.min()) * p.size / (r.max() - r.min()) / r.size

    def _bounds(self):
        p = self.phi_axis
        return (p.max() + p.min()) / 2, (p.max() - p.min())

    def add(self, roi: Roi):
        r1, r2 = roi.radius - roi.width * self.PADDING, roi.radius + roi.width * self.PADDING

        x1, x2 = self._get_r_coords(r1), self._get_r_coords(r2)
        x, y = self._get_x_y(roi, x1, x2)

        init_params, lower_bounds, upper_bounds = self._default_bounds(roi, y)
        init_curve = self.gauss(x, *init_params)

        fit = Fit(roi=roi, image_key=self.image_key, r_range=(r1, r2), x_range=(x1, x2),
                  x=x, y=y, init_curve=init_curve, init_params=init_params,
                  lower_bounds=lower_bounds, upper_bounds=upper_bounds,
                  fitted_params=[], fit_errors=[], fitting_curve=None,
                  param_names=list(self.PARAM_NAMES), fitting_method=self.METHOD)
        self.fits[roi.key] = fit

        return fit

    def update_r_range(self, fit: Fit):
        roi = fit.roi
        fit.r_range = roi.radius - roi.width * self.PADDING, roi.radius + roi.width * self.PADDING

    def update_fit(self, fit: Fit, update_bounds: bool = True):
        roi = fit.roi
        r1, r2 = fit.r_range
        x1, x2 = self._get_r_coords(r1), self._get_r_coords(r2)
        x, y = self._get_x_y(roi, x1, x2)

        init_params, lower_bounds, upper_bounds = (fit.init_params,
                                                   fit.lower_bounds,
                                                   fit.upper_bounds)

        if update_bounds and roi.movable:
            try:
                init_params, lower_bounds, upper_bounds = self._default_bounds(roi, y)
            except ValueError:
                pass
        else:
            fit.roi.radius = fit.init_params[1]
            fit.roi.width = fit.init_params[2]

        init_curve = self.gauss(x, *init_params)

        fit.update(r_range=(r1, r2), x_range=(x1, x2),
                   x=x, y=y, init_curve=init_curve, init_params=init_params,
                   lower_bounds=lower_bounds, upper_bounds=upper_bounds)

        if fit.fitted_params:
            fit.fitting_curve = self.gauss(x, *fit.fitted_params)

    def remove_fit(self, fit: Fit):
        try:
            del self.fits[fit.roi.key]
        except KeyError:
            return

    @staticmethod
    def _default_bounds(roi: Roi, y: np.ndarray):
        max_y = y.max()
        min_y = y.min()

        lower_bounds = [0, roi.radius - roi.width / 2, 0, min(0, min_y)]
        upper_bounds = [max_y, roi.radius + roi.width / 2, roi.width * 2, max_y]
        init_params = [max_y - min_y, roi.radius, roi.width, min_y]

        if roi.fitted_parameters and roi.fitted_parameters.get('method', None) == GaussianFit.METHOD:
            try:
                new_init_params = [roi.fitted_parameters[k] for k in GaussianFit.PARAM_NAMES]
                for j, (l, i, u) in enumerate(zip(lower_bounds, new_init_params, upper_bounds)):
                    if l <= i <= u:
                        init_params[j] = i
            except KeyError:
                pass

        return init_params, lower_bounds, upper_bounds

    def _get_x_y(self, roi: Roi, x1: int, x2: int):
        x = self.r_axis[x1:x2]

        if roi.type == RoiTypes.ring:
            y = self.r_profile[x1:x2]
        else:
            p1, p2 = self._get_p_coords(roi.angle - roi.angle_std / 2), \
                     self._get_p_coords(roi.angle + roi.angle_std / 2)
            y = self.polar_image[max(0, p1):min(p2, self.phi_axis.size - 1), x1:x2].sum(axis=0)

        return x, y

    def do_fit(self, fit: Fit = None) -> None:
        if fit:
            self._fit(fit)
        else:
            for fit in self.fits.values():
                self._fit(fit)

    def _fit(self, fit: Fit) -> None:
        try:
            popt, pcov = curve_fit(self.gauss, fit.x, fit.y, fit.init_params,
                                   bounds=fit.bounds)
            perr = np.sqrt(np.diag(pcov))

            fit.fitted_params = popt.tolist()
            fit.init_params = fit.fitted_params
            fit.fit_errors = perr.tolist()
            fit.fitting_curve = self.gauss(fit.x, *popt)
            fit.init_curve = self.gauss(fit.x, *fit.fitted_params)
            fit.roi.fitted_parameters = dict(zip(self.PARAM_NAMES, fit.fitted_params))
            fit.roi.fitted_parameters['method'] = self.METHOD
            fit.roi.radius = fit.roi.fitted_parameters['radius']
            fit.roi.width = fit.roi.fitted_parameters['width']

        except RuntimeError:
            pass

    @staticmethod
    def gauss(x, *p):
        amp, mu, sigma, base = p
        return amp * np.exp(- 2 * (x - mu) ** 2 / sigma ** 2) + base

    def _get_r_coords(self, r):
        return int((r - self.r_axis.min()) / self.r_delta)

    def _get_p_coords(self, p):
        return int((p - self.phi_axis.min()) / self.phi_delta)



# -*- coding: utf-8 -*-
import logging
from typing import List, Dict, Iterable, Union

from PyQt5.QtWidgets import (QWidget, QGridLayout, QPushButton,
                             QLabel, QFrame, QSplitter, QMenu)
from PyQt5.QtGui import QPen, QColor, QCursor
from PyQt5.QtCore import Qt, pyqtSlot, pyqtSignal

from .roi_widgets.roi_2d_rect_widget import Roi2DRect
from .roi_widgets.roi_1d_widget import Roi1D

from .basic_widgets import Custom1DPlot, CustomImageViewer, ParametersSlider
from ..app.fitting import Fit, GaussianFit
from ..app.rois.roi_data import RoiData, Roi, RoiTypes
from .tools import center_widget, Icon


class FitWidget(QWidget):

    sigFitApplied = pyqtSignal(object)

    def __init__(self, gaussian_fit: GaussianFit, parent=None):
        super().__init__(parent=parent)
        self.setGeometry(0, 0, 1200, 700)
        self.setWindowTitle('Fitting parameters')
        self.setWindowIcon(Icon('fit'))
        self.g_fit: GaussianFit = gaussian_fit
        self.rois: List[Roi] = [fit.roi for fit in self.g_fit.fits.values()]
        self.roi_data = RoiData(self.rois)
        self._selected_fit: Fit = None
        self._rect_widgets: dict = {}
        self._init_ui()
        self._fitted: bool = False
        center_widget(self)
        self.show()

    def _init_ui(self):
        layout = QGridLayout(self)

        self.polar_viewer = CustomImageViewer(parent=self)
        self.radial_viewer = RadialFitWidget(parent=self)
        self.fit_plot = FitPlot(parent=self)
        self.sliders_widget = SlidersWidget(self.g_fit.PARAM_NAMES, self)
        self.fit_button = QPushButton('Fit')
        self.fit_button.clicked.connect(self._fit_clicked)
        self.apply_button = QPushButton('Apply')
        self.apply_button.clicked.connect(self._apply)

        self.polar_viewer.set_data(self.g_fit.polar_image)
        self.polar_viewer.set_x_axis(self.g_fit.r_axis.min(), self.g_fit.r_axis.max())
        self.polar_viewer.set_y_axis(self.g_fit.phi_axis.min(), self.g_fit.phi_axis.max())
        self.polar_viewer.view_box.setAspectLocked(True, self.g_fit.aspect_ratio)
        self.polar_viewer.set_auto_range()
        self.radial_viewer.set_data(self.g_fit.r_axis, self.g_fit.r_profile)
        self.radial_viewer.sigUpdateFit.connect(self._update_fit)
        self.sliders_widget.sigValueChanged.connect(self._sliders_changed)

        for roi in self.rois:
            rect_widget = Roi2DRect(roi, context_menu=self._rect_context)
            rect_widget.sigSelected.connect(self._roi_selected)
            rect_widget.sigRoiMoved.connect(self._roi_moved)
            self.polar_viewer.image_plot.addItem(rect_widget)
            self._rect_widgets[roi.key] = rect_widget

        q_splitter_v = QSplitter(orientation=Qt.Horizontal, parent=self)
        q_splitter_h1 = QSplitter(orientation=Qt.Vertical, parent=self)
        q_splitter_h2 = QSplitter(orientation=Qt.Vertical, parent=self)
        q_splitter_v.addWidget(q_splitter_h1)
        q_splitter_v.addWidget(q_splitter_h2)

        q_splitter_h1.addWidget(self.polar_viewer)
        q_splitter_h1.addWidget(self.radial_viewer)
        q_splitter_h2.addWidget(self.fit_plot)
        q_splitter_h2.addWidget(self.sliders_widget)

        q_splitter_v.setSizes((800, self.width() - 800))
        q_splitter_h1.setSizes((400, self.height() - 400))
        q_splitter_h2.setSizes((400, self.height() - 400))
        layout.addWidget(q_splitter_v, 0, 0, 2, 2)

        layout.addWidget(self.fit_button, 2, 0)
        layout.addWidget(self.apply_button, 2, 1)

    def _rect_context(self, roi: Roi):
        menu = QMenu()
        menu.addAction('Delete', lambda *x, r=roi: self._delete_roi(roi))
        if roi.movable:
            menu.addAction('Fix', lambda *x, r=roi: self._fix(roi))
            menu.addAction('Fit', lambda *x, r=roi: self._fit(roi))
        else:
            menu.addAction('Unfix', lambda *x, r=roi: self._unfix(roi))
        menu.exec_(QCursor.pos())

    def _delete_roi(self, roi: Roi):
        if roi.key == self.selected_key:
            self.fit_plot.remove_fit()
            self.radial_viewer.remove_fit()
            self.sliders_widget.remove_fit()
            self._selected_fit = None
        del self.g_fit.fits[roi.key]
        self.rois.remove(roi)
        self.polar_viewer.image_plot.removeItem(self._rect_widgets[roi.key])

    @property
    def selected_key(self):
        return self._selected_fit.roi.key if self._selected_fit else None

    @pyqtSlot(name='updateFit')
    def _update_fit(self):
        self.g_fit.update_fit(self._selected_fit)
        self._rect_widgets[self.selected_key].move_roi()
        self.fit_plot.update_plot()
        self.sliders_widget.update_values()

    @pyqtSlot(int, name='roiMoved')
    def _roi_moved(self, key: int):
        if key != self.selected_key:
            self._roi_selected(key)
        roi = self._selected_fit.roi
        if roi.type == RoiTypes.ring and (roi.angle, roi.angle_std) != self.g_fit.bounds:
            roi.type = RoiTypes.segment

        self.g_fit.update_r_range(self._selected_fit)
        self.g_fit.update_fit(self._selected_fit)
        self.fit_plot.update_plot()
        self.radial_viewer.update_roi()
        self.sliders_widget.update_values()

    @pyqtSlot(name='slidersMoved')
    def _sliders_changed(self):
        self.g_fit.update_fit(self._selected_fit, update_bounds=False)
        self._rect_widgets[self.selected_key].move_roi()
        self.fit_plot.update_plot()
        self.radial_viewer.update_roi()

    @pyqtSlot(name='apply')
    def _apply(self):
        self.sigFitApplied.emit(self.g_fit)
        self.close()

    @pyqtSlot(name='fitButtonClicked')
    def _fit_clicked(self):
        if not self._fitted:
            self._fit()
            self._fitted = True
            self.fit_button.setText('Unfix')
        else:
            self._unfix()
            self._fitted = False
            self.fit_button.setText('Fit')

    def _unfix(self, roi: Union[Roi, Iterable[int]] = None):
        if isinstance(roi, Roi):
            widgets = (self._rect_widgets[roi.key],)
        elif roi:
            widgets = (self._rect_widgets[key] for key in roi)
        else:
            widgets = self._rect_widgets.values()

        for widget in widgets:
            roi = widget.roi
            roi.active = False
            roi.movable = True
            roi.fitted_parameters = None
            widget.unfix()
            if roi.key == self.selected_key:
                self.radial_viewer.roi_widget.unfix()

        # self.radial_viewer.update_roi()
        # self.fit_plot.update_plot()
        # self.sliders_widget.update_values()

    def _fix(self, roi: Union[Roi, Iterable[int]] = None):
        if isinstance(roi, Roi):
            widgets = (self._rect_widgets[roi.key],)
        elif roi:
            widgets = (self._rect_widgets[key] for key in roi)
        else:
            widgets = self._rect_widgets.values()

        for widget in widgets:
            roi = widget.roi
            roi.active = False
            roi.movable = False
            widget.fix()
            if roi.key == self.selected_key:
                self.radial_viewer.roi_widget.fix()

        # self.radial_viewer.update_roi()
        # self.fit_plot.update_plot()
        # self.sliders_widget.update_values()

    def _fit(self, roi: Union[Roi, Iterable[int]] = None):
        if isinstance(roi, Roi):
            fits = (self.g_fit.fits[roi.key], )
        elif roi:
            fits = (self.g_fit.fits[key] for key in roi)
        else:
            fits = self.g_fit.fits.values()

        keys_to_fix: List[int] = []

        for fit in fits:
            self.g_fit.do_fit(fit)
            roi_ = fit.roi
            key = roi_.key
            roi_.active = False
            if fit.fitted_params:
                keys_to_fix.append(key)

            if key == self.selected_key:
                self.fit_plot.update_plot()
                self.radial_viewer.update_roi()
                self.sliders_widget.update_values()

        self._fix(keys_to_fix)

    def _roi_selected(self, key: int):
        if key == self.selected_key:
            return
        else:
            if self._selected_fit:
                self._select_roi(self._selected_fit.roi, False)
            self._selected_fit = self.g_fit.fits[key]
            self._select_roi(self._selected_fit.roi, True)
            self._add_selected_fit()

    def _add_selected_fit(self):
        self.radial_viewer.set_fit(self._selected_fit)
        self.fit_plot.set_fit(self._selected_fit)
        self.sliders_widget.set_fit(self._selected_fit)

    def _select_roi(self, roi: Roi, active: bool):
        roi.active = active
        self._rect_widgets[roi.key].update_color()
        self.radial_viewer.plot_item.autoRange()


class SliderWithLabels(QFrame):
    def __init__(self, slider: ParametersSlider, label: QLabel, parent=None):
        super().__init__(parent=parent)
        self.slider = slider
        self.label = label
        self._init_ui()

    def _init_ui(self):
        layout = QGridLayout(self)
        layout.addWidget(self.label, 0, 0, alignment=Qt.AlignHCenter)
        layout.addWidget(self.slider, 1, 0)


class SlidersWidget(QWidget):
    sigValueChanged = pyqtSignal()
    DEFAULT_LABEL = 'lower bound; init value; upper bound'

    def __init__(self, param_names: Iterable[str], parent=None):
        super().__init__(parent=parent)
        self.fit: Fit = None
        self._init_ui(param_names)

    def _init_ui(self, param_names):
        layout = QGridLayout(self)

        self._sliders: Dict[int, ParametersSlider] = {}
        self._labels: Dict[int, QLabel] = {}

        for i, param_name in enumerate(param_names):
            layout.addWidget(QLabel(param_name), i, 0, alignment=Qt.AlignCenter)
            self._sliders[i] = sl = ParametersSlider()
            self._labels[i] = ll = QLabel(self.DEFAULT_LABEL)
            layout.addWidget(SliderWithLabels(sl, ll, self), i, 1)
            sl.sigLowerValueChanged.connect(lambda x, idx=i: self._send_value(x, 0, idx))
            sl.sigMiddleValueChanged.connect(lambda x, idx=i: self._send_value(x, 1, idx))
            sl.sigUpperValueChanged.connect(lambda x, idx=i: self._send_value(x, 2, idx))

    def set_fit(self, fit: Fit):
        self.fit = fit
        self.update_values()

    def update_values(self):
        if not self.fit:
            return
        fit = self.fit
        for idx, (l, i, u) in enumerate(zip(fit.lower_bounds, fit.init_params, fit.upper_bounds)):
            self._sliders[idx].setValues(l, i, u, new_range=(l, u))
            self._update_label(idx, l, i, u)

    def _update_label(self, idx: int, l: float, i: float, u: float):
        self._labels[idx].setText(f'lower bound = {l:0.3f}; init value = {i:0.3f}; upper bound = {u:0.3f}')

    def _send_value(self, value: float, place: int, idx: int):
        if not self.fit:
            return
        if place == 0:
            params = self.fit.lower_bounds
        elif place == 1:
            params = self.fit.init_params
        else:
            params = self.fit.upper_bounds
        params[idx] = value

        self._update_label(
            idx, self.fit.lower_bounds[idx], self.fit.init_params[idx], self.fit.upper_bounds[idx])

        self.sigValueChanged.emit()

    def remove_fit(self):
        self.fit = None
        for l in self._labels.values():
            l.setText(self.DEFAULT_LABEL)


class RadialFitWidget(Custom1DPlot):
    sigUpdateFit = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.fit: Fit = None
        self.roi_widget: Roi1D = None
        self.range_roi: Roi = None
        self.range_widget: Roi1D = None

    @pyqtSlot(object, name='addFit')
    def set_fit(self, fit: Fit):
        self.remove_fit()

        self.fit = fit
        self.roi_widget = Roi1D(fit.roi, enable_context=False)
        r1, r2 = fit.r_range

        self.range_roi = Roi(radius=(r1 + r2) / 2, width=r2 - r1, key=-1)
        self.range_widget = Roi1D(self.range_roi, enable_context=False)
        self.range_widget.set_color(QColor(255, 255, 255, 50))
        self.plot_item.addItem(self.range_widget)
        self.plot_item.addItem(self.roi_widget)

        self.range_widget.sigRoiMoved.connect(self._update_fit)
        self.roi_widget.sigRoiMoved.connect(self._update_fit)

    def update_roi(self):
        if not self.fit:
            return
        self.roi_widget.move_roi()
        r1, r2 = self.fit.r_range
        self.range_roi.radius = (r1 + r2) / 2
        self.range_roi.width = r2 - r1
        self.range_widget.move_roi()

    def _update_fit(self, key: int):
        if key == -1:
            r, w = self.range_roi.radius, self.range_roi.width
            self.fit.r_range = r - w / 2, r + w / 2
        self.sigUpdateFit.emit()

    @pyqtSlot(name='removeFit')
    def remove_fit(self):
        if self.fit:
            self.plot_item.removeItem(self.roi_widget)
            self.plot_item.removeItem(self.range_widget)
            self.fit = None
            self.roi_widget = None
            self.range_roi = None
            self.range_widget = None


class FitPlot(Custom1DPlot):
    log = logging.getLogger(__name__)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.fit: Fit = None
        self.fit_plot = self.plot_item.plot()
        self.fit_plot.setPen(self._get_fit_pen())

    def set_fit(self, fit: Fit):
        self.fit = fit
        self.update_plot()

    def update_plot(self):
        if not self.fit:
            return
        try:
            self.plot.setData(self.fit.x, self.fit.y)
            self.fit_plot.setData(self.fit.x, self.fit.init_curve)
            self.plot_item.autoRange()

        except Exception as err:
            self.log.exception(err)
            self.clear_plot()

    def remove_fit(self):
        self.fit = None
        self.clear_plot()

    def clear_plot(self):
        self.plot.clear()
        self.fit_plot.clear()

    @staticmethod
    def _get_fit_pen():
        pen = QPen(QColor('red'))
        pen.setStyle(Qt.DashDotLine)
        pen.setWidth(4)
        pen.setCapStyle(Qt.RoundCap)
        pen.setJoinStyle(Qt.RoundJoin)
        pen.setCosmetic(True)
        return pen



class SingletonMeta(type):
    _instance = None

    def __call__(cls):
        if cls._instance is None:
            cls._instance = super().__call__()
        return cls._instance


class InternalError(Exception):
    pass



# -*- coding: utf-8 -*-
import logging
from enum import Enum

import numpy as np

from PyQt5.QtWidgets import (QMainWindow, QWidget,
                             QFrame, QHBoxLayout,
                             QVBoxLayout, QPushButton,
                             QRadioButton)
from PyQt5.QtGui import QColor, QPen
from PyQt5.QtCore import Qt, pyqtSignal

from pyqtgraph import GraphicsLayoutWidget, LinearRegionItem
from .sliders import LabeledSlider
from .toolbars import BlackToolBar
from ..basic_widgets import RoundedPushButton
from ..tools import Icon, show_error
from ...app.profiles import BasicProfile

logger = logging.getLogger(__name__)


class Custom1DPlot(GraphicsLayoutWidget):
    def __init__(self, *args, parent=None):
        super().__init__(parent)
        self.plot_item = self.addPlot()
        self.plot_item.setMenuEnabled(False)
        self.plot = self.plot_item.plot(*args)
        self.set_pen()

    def set_data(self, *args):
        self.plot.setData(*args)

    def clear_plot(self):
        self.plot.clear()

    def set_x(self, x):
        x = np.array(x)
        y = self.plot.yData
        if y is None or y.shape != x.shape:
            return
        self.plot.setData(x, y)

    def set_pen(self, width: int = 3, color: str or QColor = 'white'):
        if isinstance(color, str):
            color = QColor(color)
        pen = QPen(color)
        pen.setStyle(Qt.SolidLine)
        pen.setWidth(width)
        pen.setCapStyle(Qt.RoundCap)
        pen.setJoinStyle(Qt.RoundJoin)
        pen.setCosmetic(True)
        self.plot.setPen(pen)


class Smooth1DPlot(QMainWindow):
    _MaximumSliderWidth = 200
    _MaximumSliderHeight = 30

    def __init__(self, profile: BasicProfile, parent=None):
        super().__init__(parent)
        self.profile = profile
        self.image_view = Custom1DPlot()
        self.setCentralWidget(self.image_view)
        self.__init_toolbars()

        self.profile.sigDataUpdated.connect(self.plot)

    @property
    def x(self):
        return self.profile.x

    @x.setter
    def x(self, value):
        self.profile.x = value
        self.plot()

    @property
    def y(self) -> np.ndarray or None:
        return self.profile.y

    def __init_toolbars(self):
        param_toolbar = BlackToolBar('Parameters', self)
        self.addToolBar(param_toolbar)

        self.__init_sigma_slider(param_toolbar)

    def __init_sigma_slider(self, toolbar):
        sigma_slider = LabeledSlider('𝞼', (0, 10), self.profile.sigma,
                                     decimals=2)
        sigma_slider.setMaximumWidth(self._MaximumSliderWidth)
        sigma_slider.setMaximumHeight(self._MaximumSliderHeight)
        sigma_slider.valueChanged.connect(self.update_sigma)

        self.sigma_slider = sigma_slider
        frame = QFrame()
        layout = QHBoxLayout()
        frame.setLayout(layout)
        frame.setGeometry(0, 0, self._MaximumSliderWidth, toolbar.height() * 0.9)
        layout.addWidget(sigma_slider, alignment=Qt.AlignHCenter)
        toolbar.addWidget(frame)

    def set_sigma(self, value: float):
        self.update_sigma(value)
        self.sigma_slider.set_value(value, change_bounds=True)

    def update_sigma(self, value: float):
        self.profile.set_sigma(value)
        self.plot()

    def auto_range(self):
        self.image_view.plot_item.autoRange()

    def plot(self):
        if self.x is not None and self.y is not None:
            self.image_view.set_data(self.x, self.y)
            self.auto_range()

    def clear_plot(self):
        self.image_view.clear_plot()


class PlotBC(Smooth1DPlot):

    def __init__(self, profile: BasicProfile, parent=None):
        super().__init__(profile, parent)
        self._status = BaseLineStatus.no_baseline
        self._baseline_setup_widget = None
        self.baseline_plot = None
        self.__init_roi()
        self.profile.sigDataToBeUpdated.connect(self._clear_baseline)

    @property
    def y(self):
        if self._status == BaseLineStatus.baseline_subtracted:
            return self.profile.y - self.profile.baseline
        else:
            return self.profile.y

    def update_data(self):
        self.profile.update()

    def is_shown(self, shown: bool):
        self.profile.is_shown = shown

    def __init_toolbars(self):
        super().__init_toolbars()

        baseline_toolbar = BlackToolBar('Baseline Correction')
        self.addToolBar(baseline_toolbar)

        baseline_button = RoundedPushButton(parent=baseline_toolbar, icon=Icon('baseline'),
                                            radius=30)
        baseline_button.clicked.connect(self.open_baseline_setup)
        baseline_toolbar.addWidget(baseline_button)

    def __init_roi(self):
        self._roi = LinearRegionItem()
        self._roi.hide()
        self._roi.setBrush(QColor(255, 255, 255, 50))
        self.image_view.plot_item.addItem(self.roi)

    def open_baseline_setup(self):
        if self.y is None:
            return
        self._baseline_setup_widget = setup = BaseLineSetup(
            self._status, **self.profile.get_parameters())
        if self.profile.x_range is None:
            self._set_default_bounds()

        setup.calculate_signal.connect(self._on_calculate_baseline)
        setup.subtract_signal.connect(self._on_subtracting_baseline)
        setup.restore_signal.connect(self._on_restoring_data)
        setup.close_signal.connect(self._on_closing_setup)
        setup.show()
        self.roi.show()

    def _set_default_bounds(self):
        if self.x is None:
            self.profile.x_range = (0, 1)
        else:
            self.profile.x_range = (self.x.min(), self.x.max())

    def _update_bounds(self):
        self.profile.x_range = self.roi.getRegion()

    def _set_status(self, status: 'BaseLineStatus'):
        self._status = status
        if self._baseline_setup_widget:
            self._baseline_setup_widget.set_status(status)

    def _on_calculate_baseline(self, params: dict):
        self.profile.set_parameters(**params)
        self.update_bounds()
        try:
            self.profile.update_baseline()
        except Exception as err:
            logger.exception(err)
            show_error('Failed calculating baseline. Change roi region or parameters and try again.',
                       'Baseline calculation error')
            return
        self._set_status(BaseLineStatus.baseline_calculated)
        self._plot_baseline()

    def _on_subtracting_baseline(self):
        self._remove_baseline_from_plot()
        self._set_status(BaseLineStatus.baseline_subtracted)
        self.plot()

    def _on_restoring_data(self):
        self._set_status(BaseLineStatus.baseline_restored)
        self._plot_baseline()
        self.plot()

    def _on_closing_setup(self):
        self._baseline_setup_widget = None
        self.roi.hide()
        if (self.status == BaseLineStatus.baseline_calculated or
                self.status == BaseLineStatus.baseline_restored):
            self._clear_baseline()

    def _plot_baseline(self):
        if not self.baseline_plot:
            self.baseline_plot = self.image_view.plot_item.plot()

        self.baseline_plot.setData(self.x, self.profile.baseline, pen=self._get_baseline_pen())

    @staticmethod
    def _get_baseline_pen():
        pen = QPen(QColor('red'))
        pen.setStyle(Qt.DashDotLine)
        pen.setWidth(4)
        pen.setCapStyle(Qt.RoundCap)
        pen.setJoinStyle(Qt.RoundJoin)
        pen.setCosmetic(True)
        return pen

    def _remove_baseline_from_plot(self):
        if self.baseline_plot:
            self.image_view.plot_item.removeItem(self.baseline_plot)
            self.baseline_plot = None

    def _clear_baseline(self):
        self._set_status(BaseLineStatus.no_baseline)
        self._remove_baseline_from_plot()


class BaseLineStatus(Enum):
    no_baseline = 1
    baseline_calculated = 2
    baseline_restored = 3
    baseline_subtracted = 4


class BaseLineSetup(QWidget):
    calculate_signal = pyqtSignal(dict)
    subtract_signal = pyqtSignal()
    restore_signal = pyqtSignal()
    close_signal = pyqtSignal()

    def __init__(self, status: BaseLineStatus, smoothness: float, asymmetry: float):
        super().__init__()
        self.setWindowFlag(Qt.WindowStaysOnTopHint)
        self.setWindowTitle(self.NAME)
        self.setWindowIcon(Icon('baseline'))
        self._status = None
        self.__init_ui(smoothness, asymmetry)
        self.set_status(status)

    def __init_ui(self, smoothness: float, asymmetry: float):
        layout = QVBoxLayout(self)

        self.smoothness_slider = LabeledSlider('Smoothness parameter', (1e2, 1e8), smoothness,
                                               self, Qt.Horizontal, decimals=3, scientific=True)

        self.asymmetry_slider = LabeledSlider('Asymmetry parameter', (0.001, 0.1), asymmetry,
                                              self, Qt.Horizontal, decimals=3, scientific=True)

        self.save_params_box = QRadioButton('Save as default')
        self.save_params_box.setChecked(True)
        self.calculate_button = QPushButton('Calculate baseline')
        self.calculate_button.clicked.connect(self.emit_calculate)
        self.subtract_button = QPushButton('Subtract baseline')
        self.subtract_button.clicked.connect(self.subtract_signal.emit)
        self.restore_button = QPushButton('Restore line')
        self.restore_button.clicked.connect(self.restore_signal.emit)

        layout.addWidget(self.smoothness_slider)
        layout.addWidget(self.asymmetry_slider)
        layout.addWidget(self.save_params_box)
        layout.addWidget(self.calculate_button)
        layout.addWidget(self.subtract_button)
        layout.addWidget(self.restore_button)

    def set_status(self, status: BaseLineStatus):
        if status == BaseLineStatus.no_baseline:
            self.calculate_button.setEnabled(True)
            self.subtract_button.setEnabled(False)
            self.restore_button.setEnabled(False)
        elif status == BaseLineStatus.baseline_calculated:
            self.subtract_button.setEnabled(True)
            self.restore_button.setEnabled(False)
            self.calculate_button.setEnabled(True)
        elif status == BaseLineStatus.baseline_subtracted:
            self.subtract_button.setEnabled(False)
            self.restore_button.setEnabled(True)
            self.calculate_button.setEnabled(False)
        elif status == BaseLineStatus.baseline_restored:
            self.subtract_button.setEnabled(True)
            self.restore_button.setEnabled(False)
            self.calculate_button.setEnabled(True)
        else:
            logger.error(f'Unknown status {status}')
            return
        self._status = status

    def get_params_dict(self):
        return dict(smoothness_param=self.smoothness_slider.value,
                    asymmetry_param=self.asymmetry_slider.value)

    def emit_calculate(self):
        self.calculate_signal.emit(self.get_params_dict())

    def closeEvent(self, a0) -> None:
        # if self.save_params_box.isChecked():
        #     save_config(self.NAME, self.get_params_dict())
        self.close_signal.emit()
        super().closeEvent(a0)



from pathlib import Path

from PyQt5.QtWidgets import (QMainWindow, QWidget, QSizePolicy,
                             QApplication, QShortcut, QMessageBox,
                             QFileDialog)
from PyQt5.QtCore import Qt, pyqtSlot
from PyQt5.QtGui import QKeySequence

import qdarkstyle
import qdarkgraystyle

from ..app import App
from .dock_area import AppDockArea
from .basic_widgets import ToolBar
from .tools import Icon, get_image_filepath, get_folder_filepath, save_file_dialog

from .init_window import InitWindow


class GIWAXSMainWindow(QMainWindow):
    _MinimumSize = (500, 500)

    def __init__(self, parent=None):
        super(GIWAXSMainWindow, self).__init__(parent=parent)
        self.app = App()
        self._init_toolbar()
        self._init_shortcuts()
        self._init_menubar()

        self.dock_area = AppDockArea()
        self.app.fm.sigProjectClosed.connect(self.update_window_title)

        self.setCentralWidget(self.dock_area)
        self.update_window_title()
        self.setWindowIcon(Icon('window_icon'))
        self.setMinimumSize(*self._MinimumSize)
        self.setWindowState(Qt.WindowMaximized)
        self.set_style()
        self.init_window = None

        if not self.app.fm.project_opened:
            self.open_init_window()
        else:
            self.show()

    def open_init_window(self):
        self.init_window = InitWindow(self.app.fm.recent_projects)
        self.init_window.sigOpenProject.connect(self._on_opening_project)
        self.init_window.sigExit.connect(self.close)

    def update_window_title(self):
        if self.app.fm.project_opened:
            self.setWindowTitle(f'{self.app.fm.project_name} - GIWAXS analysis')
        else:
            self.setWindowTitle('GIWAXS analysis')

    @pyqtSlot(object, name='NewProject')
    def _on_opening_project(self, path: Path):
        if self.init_window:
            self.init_window.close()
            self.init_window = None
        self.show()
        self.app.fm.open_project(path)
        self.update_window_title()

    @pyqtSlot(name='NewProjectDialog')
    def _new_project_dialog(self):
        folder = QFileDialog.getExistingDirectory(self, 'New project folder')
        if not folder:
            return
        folder = Path(folder)
        if folder.is_dir():
            self._on_opening_project(folder)

    @pyqtSlot(name='newRealTime')
    def _new_real_time(self):
        pass

    @pyqtSlot(name='addExSitu')
    def _add_file_dialog(self):
        filepath = get_image_filepath(self)
        if filepath:
            self.app.fm.add_root_path_to_project(filepath)

    def _save_as_h5(self):
        path = save_file_dialog(self, title='Save project')
        if path:
            self.app.fm.save_as_h5(path)

    def _add_folder_dialog(self):
        filepath = get_folder_filepath(self, message='Choose directory containing images or h5 files')
        if filepath:
            self.app.fm.add_root_path_to_project(filepath)

    def _init_menubar(self):
        self.menubar = self.menuBar()

        # File menu

        self.file_menu = self.menubar.addMenu('File')
        self.file_menu.addAction('New project', self._new_project_dialog)
        recent_projects_menu = self.file_menu.addMenu('Recent projects')

        for project_path in self.app.fm.recent_projects:
            recent_projects_menu.addAction(project_path.name, lambda *x, p=project_path:
                                           self._on_opening_project(p))

        # Save menu

        self.save_menu = self.file_menu.addMenu('Save project as ...')
        self.save_menu.addAction('Save as h5 file', self._save_as_h5)

        # Data menu

        self.data_menu = self.menubar.addMenu('Data')
        self.add_data_menu = self.data_menu.addMenu('Add data')

        self.add_data_menu.addAction('Add file', self._add_file_dialog)
        self.add_data_menu.addAction('Add folder', self._add_folder_dialog)

        # Preferences menu

        self.preferences = self.menubar.addMenu('Preferences')
        self.themes_menu = self.preferences.addMenu('Themes')
        themes = ['Dark', 'Gray Dark']
        # themes = CSS.list_css() + QStyleFactory.keys()
        for theme in themes:
            theme_action = self.themes_menu.addAction(theme)
            theme_action.triggered.connect(lambda *x, t=theme: self.set_style(t))

    def _init_shortcuts(self):
        self.copy_shortcut = QShortcut(QKeySequence("Ctrl+C"), self)
        self.copy_shortcut.activated.connect(lambda *x: self.app.roi_dict.copy_rois('selected'))

        self.paste_shortcut = QShortcut(QKeySequence("Ctrl+V"), self)
        self.paste_shortcut.activated.connect(lambda *x: self.app.roi_dict.paste_rois())

        self.select_shortcut = QShortcut(QKeySequence("Ctrl+A"), self)
        self.select_shortcut.activated.connect(lambda *x: self.app.roi_dict.select_all())

        self.delete_shortcut = QShortcut(QKeySequence('Del'), self)
        self.delete_shortcut.activated.connect(lambda *x: self.app.roi_dict.delete_selected_roi())

        self.fit_shortcut = QShortcut(QKeySequence('Ctrl+F'), self)
        self.fit_shortcut.activated.connect(lambda *x: self.app.roi_dict.open_fit_rois(True))

        def raise_err(*_):
            raise ValueError('Congratulations, you found an error combination used for testing! '
                             'Your project will be deleted in 3 seconds...')

        self.raise_shortcut = QShortcut(QKeySequence('Ctrl+R'), self)
        self.raise_shortcut.activated.connect(raise_err)

    def _init_toolbar(self):

        docks_toolbar = ToolBar('Docks', self)
        self.addToolBar(docks_toolbar)

        control_widget = docks_toolbar.addAction(Icon('folder'), 'File Manager')
        control_widget.triggered.connect(lambda: self.dock_area.show_hide_docks('file_widget'))

        radial_profile = docks_toolbar.addAction(Icon('radial_profile'), 'Radial profile')
        radial_profile.triggered.connect(lambda: self.dock_area.show_hide_docks('radial_profile'))

        angular_profile = docks_toolbar.addAction(Icon('angular_profile'), 'Angular profile')
        angular_profile.triggered.connect(lambda: self.dock_area.show_hide_docks('angular_profile'))

        interpolation = docks_toolbar.addAction(Icon('interpolate'), 'Polar Viewer')
        interpolation.triggered.connect(lambda: self.dock_area.show_hide_docks('polar'))

        self.gen_toolbar = ToolBar('General')
        self.addToolBar(self.gen_toolbar)
        spacer_widget = QWidget()
        spacer_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        spacer_widget.setVisible(True)
        self.gen_toolbar.addWidget(spacer_widget)

        self.fullscreen_action = self.gen_toolbar.addAction(Icon('tofullscreen'), 'Full screen')
        self.fullscreen_action.triggered.connect(self._on_fullscreen_changed)

    def _on_fullscreen_changed(self):
        if self.isFullScreen():
            self.setWindowState(Qt.WindowMaximized)
            self.fullscreen_action.setIcon(Icon('tofullscreen'))
        else:
            self.setWindowState(Qt.WindowFullScreen)
            self.fullscreen_action.setIcon(Icon('fromfullscreen'))

    def set_style(self, name: str = 'Gray Dark'):
        qapp = QApplication.instance()
        if not qapp:
            raise RuntimeError('No running application found.')
        if name == 'Dark':
            qapp.setStyleSheet('')
            qapp.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
        elif name == 'Gray Dark':
            qapp.setStyleSheet('')
            qapp.setStyleSheet(qdarkgraystyle.load_stylesheet_pyqt5())
        else:
            return

        self.app.fm.config['style'] = name

    def close(self):
        self.app.close()
        qapp = QApplication.instance()
        qapp.closeAllWindows()

    def closeEvent(self, a0) -> None:
        reply = QMessageBox.question(self, 'Message',
                                     "Are you sure to quit? The project will be saved.",
                                     QMessageBox.Yes |
                                     QMessageBox.No, QMessageBox.No)

        if reply == QMessageBox.Yes:
            a0.accept()
            self.close()
        else:
            a0.ignore()



import logging
from pathlib import Path
from typing import List

from PyQt5.QtWidgets import (QGraphicsColorizeEffect, QLineEdit,
                             QWidget, QApplication, QMessageBox,
                             QFileDialog)
from PyQt5.QtCore import QPropertyAnimation, Qt
from PyQt5.QtGui import QColor, QIcon

from ..app.file_manager import GLOB_IMAGE_FORMATS

ICON_PATH: Path = Path(__file__).parents[1] / 'static' / 'icons'
CSS_PATH: Path = Path(__file__).parents[1] / 'static' / 'css'

logger = logging.getLogger(__name__)


def save_file_dialog(parent, title: str = 'Save File', file_format: str = 'H5 file (*.h5)'
                     ):
    options = QFileDialog.Options()
    options |= QFileDialog.DontUseNativeDialog
    filename, _ = QFileDialog.getSaveFileName(parent, title, '', file_format, options=options)
    if filename:
        return Path(filename)


def get_image_filepath(parent, message: str = 'Open image') -> Path or None:
    options = QFileDialog.Options()
    options |= QFileDialog.DontUseNativeDialog
    filepath, _ = QFileDialog.getOpenFileName(
        parent, message, '',
        GLOB_IMAGE_FORMATS, options=options)
    if filepath:
        return Path(filepath)


def get_folder_filepath(parent, message: str, *, show_files: bool = True) -> Path or None:
    options = QFileDialog.DontResolveSymlinks | QFileDialog.DontUseNativeDialog
    if not show_files:
        options |= QFileDialog.ShowDirsOnly

    folder_path = QFileDialog.getExistingDirectory(
        parent, message, '',
        options=options)
    if folder_path:
        return Path(folder_path)


def show_error(err: str, *, error_title: str = 'Internal Error', info_text: str = ''):
    logger.info(f'Error message shown: {error_title} - {err} {info_text}.')

    mb = QMessageBox()
    mb.setIcon(QMessageBox.Critical)
    mb.setWindowTitle(error_title)
    mb.setWindowIcon(Icon('error'))
    mb.setText(err)
    if info_text:
        mb.setInformativeText(info_text)
    mb.exec_()


class Icon(QIcon):
    def __init__(self, name: str):
        if name.find('.') == -1:
            name += '.png'
        name = str(ICON_PATH / name)
        QIcon.__init__(self, name)


class CSS(object):
    default = 'Dark Grey'

    @staticmethod
    def get_css(name: str) -> str or None:
        if '.css' not in name:
            name += '.css'
        try:
            with open(str((CSS_PATH / name).resolve()), 'r') as f:
                return f.read()
        except FileNotFoundError:
            return
        except Exception as err:
            logger.exception(err)
            return

    @staticmethod
    def list_css() -> List[str]:
        return [p.stem for p in CSS_PATH.glob('*.css')]


def center_widget(widget):
    frame_gm = widget.frameGeometry()
    screen = QApplication.desktop().screenNumber(
        QApplication.desktop().cursor().pos())
    center_point = QApplication.desktop().screenGeometry(
        screen).center()
    frame_gm.moveCenter(center_point)
    widget.move(frame_gm.topLeft())


def validate_scientific_value(q_line_edit: QLineEdit,
                              data_type: type or None = float,
                              empty_possible: bool = False,
                              additional_conditions: tuple = ()):
    text_value = q_line_edit.text().replace(',', '.')
    if data_type is None:
        return text_value
    try:
        value = data_type(text_value)
    except ValueError:
        if not empty_possible:
            color_animation(q_line_edit)
        return
    for condition in additional_conditions:
        if not condition(value):
            return
    return value


def color_animation(widget: QWidget, color=Qt.red):
    effect = QGraphicsColorizeEffect(widget)
    widget.setGraphicsEffect(effect)

    widget.animation = QPropertyAnimation(effect, b'color')

    widget.animation.setStartValue(QColor(color))
    widget.animation.setEndValue(QColor(Qt.black))

    widget.animation.setLoopCount(1)
    widget.animation.setDuration(1500)
    widget.animation.start()
